import calendar
from datetime import date, datetime, timedelta
from typing import Union

import dateparser
from django.utils import timezone

from jnt_django_toolbox.consts.time import (
    SECONDS_PER_DAY,
    SECONDS_PER_HOUR,
    SECONDS_PER_MINUTE,
)

Number = Union[int, float]


def date2datetime(value: date) -> datetime:  # noqa: WPS110
    """Converts date to datetime."""
    return datetime.combine(value, datetime.min.time())


def begin_of_week(target: date, first_weekday: int = calendar.MONDAY) -> date:
    """Get begin of week."""
    days_passed = target.weekday() - first_weekday

    if days_passed < 0:
        days_passed += 7

    return target - timedelta(days=days_passed)


def humanize_time(total_seconds: Number) -> str:
    """Provides human friendly representation for seconds."""
    if not isinstance(total_seconds, (int, float)):
        raise ValueError("Seconds should be a number")

    if total_seconds == 0:
        return "0s"

    items = []

    time_units = (
        ("d", SECONDS_PER_DAY),
        ("h", SECONDS_PER_HOUR),
        ("m", SECONDS_PER_MINUTE),
        ("s", 1),
    )

    for unit, sec_in_unit in time_units:
        val = total_seconds // sec_in_unit
        if not val:
            continue

        items.append("{0}{1}".format(int(val), unit))
        total_seconds -= sec_in_unit * val
        if not total_seconds:
            break

    return " ".join(items)


epoch = datetime.utcfromtimestamp(0)


def unix_time_seconds(dt):
    """Get unix time from datetime."""
    return (dt.replace(tzinfo=None) - epoch).total_seconds()


def parse_human_date(date_str: str):
    """Parse human presented date string."""
    return timezone.make_aware(dateparser.parse(date_str))



import time

from jnt_django_toolbox.profiling.profilers.base import BaseProfiler


class RequestsProfiler(BaseProfiler):
    """Request profiler."""

    def __init__(self, header_prefix: str = "app_req"):
        """Initializing."""
        self._header_prefix = header_prefix

    def before_request(self, request, stack):
        """Start capturing requests."""
        self._start_time = time.time()

    def after_request(self, request, response):
        """Add profiling info to response."""
        response[self._header("time")] = round(
            time.time() - self._start_time,
            3,
        )

    def _header(self, name: str) -> str:
        return "{0}_{1}".format(self._header_prefix, name)



from django.db import models

from jnt_django_toolbox.forms.fields import EnumChoiceField


class EnumField(models.TextField):
    """
    Wrapper around django choices.

    Allow better using of django 3 choices.
    """

    def __init__(self, enum, *args, **kwargs):
        """Initializer."""
        self.enum = enum
        kwargs["choices"] = enum.choices

        super().__init__(*args, **kwargs)

    def deconstruct(self):
        """Deconstruct field for clone."""
        name, path, args, kwargs = super().deconstruct()
        kwargs["enum"] = self.enum

        return name, path, args, kwargs

    def formfield(self, **kwargs):
        """Build form field."""
        kwargs["choices_form_class"] = EnumChoiceField
        form_field = super().formfield(**kwargs)
        form_field.enum = self.enum

        return form_field



import time
from collections import defaultdict
from contextlib import ExitStack

from django.core.cache import cache
from django.http import HttpRequest, HttpResponse

from jnt_django_toolbox.profiling.profilers.base import BaseProfiler

_CACHE_METHODS = (
    "add",
    "get",
    "set",
    "delete",
    "clear",
    "has_key",
    "get_many",
    "set_many",
)


class _CaptureCacheCallsContext:
    def __init__(self, cache_provider):
        self._cache = cache_provider

    def __enter__(self):
        self.stats = defaultdict(list)
        self.original_methods = {}

        for method in _CACHE_METHODS:
            self.original_methods[method] = getattr(cache, method)
            setattr(
                cache,
                method,
                self._track_call(self.original_methods[method]),
            )

    def __exit__(self, exc_type, exc_value, traceback):
        for method in _CACHE_METHODS:
            setattr(cache, method, self.original_methods[method])

    def _track_call(self, actual_cache_call):
        def wrapper(*args, **kwargs):  # noqa:WPS430
            start = time.time()
            return_value = actual_cache_call(*args, **kwargs)
            duration = time.time() - start

            self.stats[actual_cache_call.__name__].append(duration)
            return return_value

        return wrapper


class CacheCallsProfiler(BaseProfiler):
    """Cache calls count profiler."""

    def __init__(self, header_prefix: str = "app_cache"):
        """Initializing."""
        self._header_prefix = header_prefix

    def before_request(self, request: HttpRequest, stack: ExitStack):
        """Start capturing cache calls."""
        self._context = _CaptureCacheCallsContext(cache)
        stack.enter_context(self._context)

    def after_request(self, request: HttpRequest, response: HttpResponse):
        """Add profiling info to response."""
        stats = self._context.stats

        response["{0}_time".format(self._header_prefix)] = round(
            sum(sum(duration) for duration in stats.values()),
            3,
        )

        for name, durations in sorted(stats.items()):
            header = "{0}_{1}_count".format(self._header_prefix, name)
            response[header] = len(durations)



from django.core.cache import cache
from django.core.management.commands import migrate

from jnt_django_toolbox.context_managers.global_lock import (
    build_global_cache_key,
)
from jnt_django_toolbox.management.commands.migrate_concurrently import (
    MIGRATE_LOCK_NAME,
)


class Command(migrate.Command):
    """Reset cache lock for preventing parallel migrations."""

    def handle(self, *args, **options):  # noqa: WPS110
        """Command body."""
        cache.delete(build_global_cache_key(MIGRATE_LOCK_NAME))
        self.stdout.write(self.style.SUCCESS("Migrate lock was reseted"))
